<?php

declare( strict_types=1 );

namespace Wikibase\Client\Store\Sql;

use InvalidArgumentException;
use Onoi\MessageReporter\MessageReporter;
use Onoi\MessageReporter\NullMessageReporter;
use Wikibase\Client\NamespaceChecker;
use Wikibase\Lib\Rdbms\ClientDomainDb;
use Wikimedia\Rdbms\SessionConsistentConnectionManager;

/**
 * Maintenance helper which adds the "unexpectedUnconnectedPage" page property
 * to all relevant pages.
 *
 * @license GPL-2.0-or-later
 * @author Marius Hoch <mail@mariushoch.de>
 */
class UnexpectedUnconnectedPagePrimer {

	/**
	 * @var SessionConsistentConnectionManager
	 */
	private $localConnectionManager;

	/**
	 * @var ClientDomainDb
	 */
	private $clientDb;

	/**
	 * @var NamespaceChecker
	 */
	private $namespaceChecker;

	/**
	 * @var int
	 */
	private $batchSize;

	/**
	 * @var int
	 */
	private $batchSizeSelectMultiplicator = 1000;

	/**
	 * @var MessageReporter
	 */
	private $progressReporter;

	/**
	 * @var int
	 */
	private $position = 0;

	/**
	 * @var int|null
	 */
	private $maxPageId = null;

	/**
	 * @param ClientDomainDb $clientDb
	 * @param NamespaceChecker $namespaceChecker
	 * @param int $batchSize
	 *
	 * @throws InvalidArgumentException
	 */
	public function __construct(
		ClientDomainDb $clientDb,
		NamespaceChecker $namespaceChecker,
		int $batchSize = 1000
	) {
		if ( $batchSize < 1 ) {
			throw new InvalidArgumentException( '$batchSize must be an integer >= 1' );
		}
		$this->clientDb = $clientDb;
		$this->localConnectionManager = $clientDb->sessionConsistentConnections();
		$this->batchSize = $batchSize;
		$this->namespaceChecker = $namespaceChecker;

		$this->progressReporter = new NullMessageReporter();
	}

	/**
	 * Set the batch size multiplicator used to determine the number of page ids to
	 * probe when selecting for a batch.
	 *
	 * @param int $batchSizeSelectMultiplicator
	 */
	public function setBatchSizeSelectMultiplicator( int $batchSizeSelectMultiplicator ): void {
		$this->batchSizeSelectMultiplicator = $batchSizeSelectMultiplicator;
	}

	/**
	 * Set the page ID at which to start processing (inclusive).
	 *
	 * @param int|null $minPageId The page ID, or null for no offset (default).
	 */
	public function setMinPageId( ?int $minPageId ): void {
		$this->position = ( $minPageId ?: 1 ) - 1;
	}

	/**
	 * Set the page ID at which to stop processing (inclusive).
	 * This is a rough measure – up to batch size further pages may be processed.
	 *
	 * @param int|null $maxPageId The page ID, or null for no limit (default).
	 */
	public function setMaxPageId( ?int $maxPageId ): void {
		$this->maxPageId = $maxPageId;
	}

	/**
	 * Add the "unexpectedUnconnectedPage" page prop for all relevant pages.
	 */
	public function insertPageProp(): void {
		$highestPageId = $this->getMaximumPageIdToCheck();
		$maxPageId = min( $this->maxPageId ?: PHP_INT_MAX, $highestPageId );

		while ( $this->position < $maxPageId ) {
			$count = $this->processUnexpectedUnconnectedBatch();

			if ( $count > 0 ) {
				$this->progressReporter->reportMessage(
					'Added the "unexpectedUnconnectedPage" page property for ' . $count . ' pages, ' .
					'up to page ID ' . $this->position . ' (inclusive).'
				);
				$this->clientDb->replication()->wait();
			}
		}

		if ( $this->position >= $highestPageId ) {
			$this->progressReporter->reportMessage( 'Done!' );
		} // otherwise, need to re-run with new min/max page ID
	}

	/**
	 * @return int Number of changes performed.
	 */
	private function processUnexpectedUnconnectedBatch() {
		$pages = $this->getUnexpectedUnconnectedBatch();

		if ( !$pages ) {
			return 0;
		}

		$count = $this->insertUnexpectedUnconnectedBatch( $pages );
		return $count;
	}

	/**
	 * @param int[][] $pages Page id, page namespace pairs
	 *
	 * @return int The number of rows inserted.
	 */
	private function insertUnexpectedUnconnectedBatch( array $pages ): int {
		$rows = $this->makeUnexpectedUnconnectedRows( $pages );

		$dbw = $this->localConnectionManager->getWriteConnectionRef();
		$dbw->startAtomic( __METHOD__ );

		$dbw->insert(
			'page_props',
			$rows,
			__METHOD__,
			[
				'IGNORE'
			]
		);

		$count = $dbw->affectedRows();
		$dbw->endAtomic( __METHOD__ );

		return $count;
	}

	/**
	 * Returns a list of rows for insertion, using DatabaseBase's multi-row insert mechanism.
	 *
	 * @param int[][] $pages Page id, page namespace pairs
	 *
	 * @return array[] rows
	 */
	private function makeUnexpectedUnconnectedRows( array $pages ): array {
		$rows = [];

		foreach ( $pages as $page ) {
			$rows[] = [
				'pp_page' => $page[0],
				'pp_propname' => 'unexpectedUnconnectedPage',
				'pp_value' => $page[1],
				'pp_sortkey' => $page[1],
			];
		}

		return $rows;
	}

	/**
	 * @return int[][] Page id, page namespace pairs
	 */
	private function getUnexpectedUnconnectedBatch(): array {
		$dbr = $this->localConnectionManager->getReadConnectionRef();

		$lastPosition = $this->position + $this->batchSize * $this->batchSizeSelectMultiplicator;
		$result = $dbr->select(
			[ 'page', 'page_props' ],
			[ 'page_id', 'page_namespace' ],
			[
				'pp_propname' => null,
				'page_namespace' => $this->namespaceChecker->getWikibaseNamespaces(),
				'page_is_redirect' => 0,
				'page_id > ' . $this->position,
				'page_id <= ' . $lastPosition,
			],
			__METHOD__,
			[
				'LIMIT' => $this->batchSize,
			],
			[
				'page_props' => [
					'LEFT JOIN',
					[ 'page_id = pp_page', 'pp_propname' => [ 'wikibase_item', 'expectedUnconnectedPage' ] ]
				]
			]
		);
		$pages = [];
		foreach ( $result as $row ) {
			$pages[] = [ $row->page_id, $row->page_namespace ];
		}

		if ( count( $pages ) < $this->batchSize ) {
			$this->position = $lastPosition;
		} else {
			$this->position = intval( end( $pages )[0] );
		}

		return $pages;
	}

	/**
	 * @return int The largest page id we need to bother looking for.
	 */
	private function getMaximumPageIdToCheck(): int {
		// Pages added now are fine anyway, as we assume the new page prop to be active when this
		// script is run.
		return (int)$this->localConnectionManager->getReadConnectionRef()->selectField(
			'page',
			'MAX(page_id)',
			[],
			__METHOD__
		);
	}

}
